/*
 * Sixth 3D engine. Copyright ©2012-2019, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 *
 */

package eu.svjatoslav.sixth.e3d.renderer.raster;

public final class Color {

    public static final Color RED = new Color(255, 0, 0, 255);
    public static final Color GREEN = new Color(0, 255, 0, 255);
    public static final Color BLUE = new Color(0, 0, 255, 255);
    public static final Color YELLOW = new Color(255, 255, 0, 255);
    public static final Color WHITE = new Color(255, 255, 255, 255);
    public static final Color BLACK = new Color(0, 0, 0, 255);
    public static final Color PURPLE = new Color(255, 0, 255, 255);
    public static final Color TRANSPARENT = new Color(0, 0, 0, 0);

    public final int r, g, b;

    /**
     * <pre>
     * 255 is opaque.
     * 0 is transparent.
     * </pre>
     */
    public int a;

    public Color(final Color parentColor) {
        r = parentColor.r;
        g = parentColor.g;
        b = parentColor.b;
        a = parentColor.a;
    }

    public Color(final double r, final double g, final double b, final double a) {
        this.r = ensureByteLimit((int) (r * 255d));
        this.g = ensureByteLimit((int) (g * 255d));
        this.b = ensureByteLimit((int) (b * 255d));
        this.a = ensureByteLimit((int) (a * 255d));
    }

    /**
     * <pre>
     *     Supported formats are:
     *
     *     RGB
     *     RGBA
     *     RRGGBB
     *     RRGGBBAA
     * </pre>
     */
    public Color(String colorHexCode) {
        switch (colorHexCode.length()) {
            case 3:
                r = parseHexSegment(colorHexCode, 0, 1) * 16;
                g = parseHexSegment(colorHexCode, 1, 1) * 16;
                b = parseHexSegment(colorHexCode, 2, 1) * 16;
                a = 255;
                return;

            case 4:
                r = parseHexSegment(colorHexCode, 0, 1) * 16;
                g = parseHexSegment(colorHexCode, 1, 1) * 16;
                b = parseHexSegment(colorHexCode, 2, 1) * 16;
                a = parseHexSegment(colorHexCode, 3, 1) * 16;
                return;

            case 6:
                r = parseHexSegment(colorHexCode, 0, 2);
                g = parseHexSegment(colorHexCode, 2, 2);
                b = parseHexSegment(colorHexCode, 4, 2);
                a = 255;
                return;

            case 8:
                r = parseHexSegment(colorHexCode, 0, 2);
                g = parseHexSegment(colorHexCode, 2, 2);
                b = parseHexSegment(colorHexCode, 4, 2);
                a = parseHexSegment(colorHexCode, 6, 2);
                return;
            default:
                throw new IllegalArgumentException("Unsupported color code: " + colorHexCode);
        }
    }

    public Color(final int rgb) {
        r = (rgb & 0xFF0000) >> 16;
        g = (rgb & 0xFF00) >> 8;
        b = rgb & 0xFF;
        a = 255;
    }

    public Color(final int r, final int g, final int b) {
        this(r, g, b, 255);
    }

    public Color(final int r, final int g, final int b, final int a) {
        this.r = ensureByteLimit(r);
        this.g = ensureByteLimit(g);
        this.b = ensureByteLimit(b);
        this.a = ensureByteLimit(a);
    }

    private int parseHexSegment(String hexString, int start, int length) {
        return Integer.parseInt(hexString.substring(start, start + length), 16);
    }

    /**
     * Ensure that color values are within allowed limits of 0 to 255.
     */
    private int ensureByteLimit(final int value) {
        if (value < 0)
            return 0;

        if (value > 255)
            return 255;

        return value;
    }

    public boolean isTransparent() {
        return a == 0;
    }

    public java.awt.Color toAwtColor() {
        return new java.awt.Color(r, g, b, a);
    }

    public int toInt() {
        return toAwtColor().getRGB();
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        Color color = (Color) o;

        if (r != color.r) return false;
        if (g != color.g) return false;
        if (b != color.b) return false;
        return a == color.a;
    }

    @Override
    public int hashCode() {
        int result = r;
        result = 31 * result + g;
        result = 31 * result + b;
        result = 31 * result + a;
        return result;
    }
}
