/*
 * Sixth 3D engine. Copyright ©2012-2019, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 *
 */


package eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic.line;

import eu.svjatoslav.sixth.e3d.geometry.Point2D;
import eu.svjatoslav.sixth.e3d.geometry.Point3D;
import eu.svjatoslav.sixth.e3d.gui.RenderingContext;
import eu.svjatoslav.sixth.e3d.renderer.raster.Color;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.AbstractCoordinateShape;

public class Line extends AbstractCoordinateShape {

    private static final double MINIMUM_WIDTH_THRESHOLD = 1;

    private static final double LINE_WIDTH_MULTIPLIER = 0.2d;

    public final double width;
    final LineInterpolator[] li = new LineInterpolator[4];
    public Color color;

    public Line(final Line parentLine) {
        this(parentLine.coordinates[0].coordinate.clone(),
                parentLine.coordinates[1].coordinate.clone(),
                new Color(parentLine.color), parentLine.width);
    }

    public Line(final Point3D point1, final Point3D point2, final Color color,
                final double width) {

        super(point1, point2);

        this.color = color;
        this.width = width;

        for (int i = 0; i < li.length; i++)
            li[i] = new LineInterpolator();

    }

    private void drawHorizontalLine(final LineInterpolator line1,
                                    final LineInterpolator line2, final int y,
                                    final RenderingContext renderBuffer) {

        int x1 = line1.getX(y);
        int x2 = line2.getX(y);

        double d1 = line1.getD();
        double d2 = line2.getD();

        if (x1 > x2) {
            final int tmp = x1;
            x1 = x2;
            x2 = tmp;

            final double tmp2 = d1;
            d1 = d2;
            d2 = tmp2;
        }

        final int unclippedWidth = x2 - x1;
        final double dinc = (d2 - d1) / unclippedWidth;

        if (x1 < 0) {
            d1 += (dinc * (-x1));
            x1 = 0;
        }

        if (x2 >= renderBuffer.width)
            x2 = renderBuffer.width - 1;

        final int drawnWidth = x2 - x1;

        int offset = ((y * renderBuffer.width) + x1) * 4;
        final byte[] offSreenBufferBytes = renderBuffer.pixels;

        final int lineAlpha = color.a;

        final int colorB = color.b;
        final int colorG = color.g;
        final int colorR = color.r;

        for (int i = 0; i < drawnWidth; i++) {

            final double alphaMultiplier = 1d - Math.abs(d1);

            final int realLineAlpha = (int) (lineAlpha * alphaMultiplier);
            final int backgroundAlpha = 255 - realLineAlpha;

            offSreenBufferBytes[offset] = (byte) 255;
            offset++;
            offSreenBufferBytes[offset] = (byte) ((((offSreenBufferBytes[offset] & 0xff) * backgroundAlpha) + (colorB * realLineAlpha)) / 256);
            offset++;
            offSreenBufferBytes[offset] = (byte) ((((offSreenBufferBytes[offset] & 0xff) * backgroundAlpha) + (colorG * realLineAlpha)) / 256);
            offset++;
            offSreenBufferBytes[offset] = (byte) ((((offSreenBufferBytes[offset] & 0xff) * backgroundAlpha) + (colorR * realLineAlpha)) / 256);
            offset++;

            d1 += dinc;
        }

    }

    private void drawSinglePixelHorizontalLine(final RenderingContext buffer,
                                               final int alpha) {

        final Point2D onScreenPoint1 = coordinates[0].onScreenCoordinate;
        final Point2D onScreenPoint2 = coordinates[1].onScreenCoordinate;

        int xStart = (int) onScreenPoint1.x;
        int xEnd = (int) onScreenPoint2.x;

        int lineHeight;
        int yBase;

        if (xStart > xEnd) {
            final int tmp = xStart;
            xStart = xEnd;
            xEnd = tmp;
            lineHeight = (int) (onScreenPoint1.y - onScreenPoint2.y);
            yBase = (int) onScreenPoint2.y;
        } else {
            yBase = (int) onScreenPoint1.y;
            lineHeight = (int) (onScreenPoint2.y - onScreenPoint1.y);
        }

        final int lineWidth = xEnd - xStart;
        if (lineWidth == 0)
            return;

        final byte[] offSreenBufferBytes = buffer.pixels;
        final int backgroundAlpha = 255 - alpha;

        final int blueWithAlpha = color.b * alpha;
        final int greenWithAplha = color.g * alpha;
        final int redWithAlpha = color.r * alpha;

        for (int relativeX = 0; relativeX <= lineWidth; relativeX++) {
            final int x = xStart + relativeX;

            if ((x >= 0) && (x < buffer.width)) {

                final int y = yBase + ((relativeX * lineHeight) / lineWidth);
                if ((y >= 0) && (y < buffer.height)) {
                    int ramOffset = ((y * buffer.width) + x) * 4;

                    offSreenBufferBytes[ramOffset] = (byte) 255;
                    ramOffset++;
                    offSreenBufferBytes[ramOffset] = (byte) ((((offSreenBufferBytes[ramOffset] & 0xff) * backgroundAlpha) + blueWithAlpha) / 256);
                    ramOffset++;
                    offSreenBufferBytes[ramOffset] = (byte) ((((offSreenBufferBytes[ramOffset] & 0xff) * backgroundAlpha) + greenWithAplha) / 256);
                    ramOffset++;
                    offSreenBufferBytes[ramOffset] = (byte) ((((offSreenBufferBytes[ramOffset] & 0xff) * backgroundAlpha) + redWithAlpha) / 256);
                }
            }
        }

    }

    private void drawSinglePixelVerticalLine(final RenderingContext buffer,
                                             final int alpha) {

        final Point2D onScreenPoint1 = coordinates[0].onScreenCoordinate;
        final Point2D onScreenPoint2 = coordinates[1].onScreenCoordinate;

        int yStart = (int) onScreenPoint1.y;
        int yEnd = (int) onScreenPoint2.y;

        int lineWidth;
        int xBase;

        if (yStart > yEnd) {
            final int tmp = yStart;
            yStart = yEnd;
            yEnd = tmp;
            lineWidth = (int) (onScreenPoint1.x - onScreenPoint2.x);
            xBase = (int) onScreenPoint2.x;
        } else {
            xBase = (int) onScreenPoint1.x;
            lineWidth = (int) (onScreenPoint2.x - onScreenPoint1.x);
        }

        final int lineHeight = yEnd - yStart;
        if (lineHeight == 0)
            return;

        final byte[] offScreenBufferBytes = buffer.pixels;
        final int backgroundAlpha = 255 - alpha;

        final int blueWithAlpha = color.b * alpha;
        final int greenWithAlpha = color.g * alpha;
        final int redWithAlpha = color.r * alpha;

        for (int relativeY = 0; relativeY <= lineHeight; relativeY++) {
            final int y = yStart + relativeY;

            if ((y >= 0) && (y < buffer.height)) {

                final int x = xBase + ((relativeY * lineWidth) / lineHeight);
                if ((x >= 0) && (x < buffer.width)) {
                    int ramOffset = ((y * buffer.width) + x) * 4;

                    offScreenBufferBytes[ramOffset] = (byte) 255;
                    ramOffset++;
                    offScreenBufferBytes[ramOffset] = (byte) ((((offScreenBufferBytes[ramOffset] & 0xff) * backgroundAlpha) + blueWithAlpha) / 256);
                    ramOffset++;
                    offScreenBufferBytes[ramOffset] = (byte) ((((offScreenBufferBytes[ramOffset] & 0xff) * backgroundAlpha) + greenWithAlpha) / 256);
                    ramOffset++;
                    offScreenBufferBytes[ramOffset] = (byte) ((((offScreenBufferBytes[ramOffset] & 0xff) * backgroundAlpha) + redWithAlpha) / 256);
                }
            }
        }
    }

    private int getLineInterpolator(final int startPointer, final int y) {

        for (int i = startPointer; i < li.length; i++)
            if (li[i].containsY(y))
                return i;
        return -1;
    }

    @Override
    public void paint(final RenderingContext buffer) {

        final Point2D onScreenPoint1 = coordinates[0].onScreenCoordinate;
        final Point2D onScreenPoint2 = coordinates[1].onScreenCoordinate;

        final double xp = onScreenPoint2.x - onScreenPoint1.x;
        final double yp = onScreenPoint2.y - onScreenPoint1.y;

        final double point1radius = (buffer.width * LINE_WIDTH_MULTIPLIER * width)
                / coordinates[0].transformedCoordinate.z;
        final double point2radius = (buffer.width * LINE_WIDTH_MULTIPLIER * width)
                / coordinates[1].transformedCoordinate.z;

        if ((point1radius < MINIMUM_WIDTH_THRESHOLD)
                || (point2radius < MINIMUM_WIDTH_THRESHOLD)) {

            double averageRadius = (point1radius + point2radius) / 2;

            if (averageRadius > 1)
                averageRadius = 1;

            final int alpha = (int) (color.a * averageRadius);
            if (alpha < 2)
                return;

            if (Math.abs(xp) > Math.abs(yp))
                drawSinglePixelHorizontalLine(buffer, alpha);
            else
                drawSinglePixelVerticalLine(buffer, alpha);
            return;
        }

        final double lineLength = Math.sqrt((xp * xp) + (yp * yp));

        final double yinc1 = (point1radius * xp) / lineLength;
        final double yinc2 = (point2radius * xp) / lineLength;

        final double xdec1 = (point1radius * yp) / lineLength;
        final double xdec2 = (point2radius * yp) / lineLength;

        final double p1x1 = onScreenPoint1.x - xdec1;
        final double p1y1 = onScreenPoint1.y + yinc1;

        final double p1x2 = onScreenPoint1.x + xdec1;
        final double p1y2 = onScreenPoint1.y - yinc1;

        final double p2x1 = onScreenPoint2.x - xdec2;
        final double p2y1 = onScreenPoint2.y + yinc2;

        final double p2x2 = onScreenPoint2.x + xdec2;
        final double p2y2 = onScreenPoint2.y - yinc2;

        li[0].setPoints(p1x1, p1y1, 1d, p2x1, p2y1, 1d);
        li[1].setPoints(p1x2, p1y2, -1d, p2x2, p2y2, -1d);

        li[2].setPoints(p1x1, p1y1, 1d, p1x2, p1y2, -1d);
        li[3].setPoints(p2x1, p2y1, 1d, p2x2, p2y2, -1d);

        double ymin = p1y1;
        if (p1y2 < ymin)
            ymin = p1y2;
        if (p2y1 < ymin)
            ymin = p2y1;
        if (p2y2 < ymin)
            ymin = p2y2;
        if (ymin < 0)
            ymin = 0;

        double ymax = p1y1;
        if (p1y2 > ymax)
            ymax = p1y2;
        if (p2y1 > ymax)
            ymax = p2y1;
        if (p2y2 > ymax)
            ymax = p2y2;
        if (ymax >= buffer.height)
            ymax = buffer.height - 1;

        for (int y = (int) ymin; y <= ymax; y++) {
            final int li1 = getLineInterpolator(0, y);
            if (li1 != -1) {
                final int li2 = getLineInterpolator(li1 + 1, y);
                if (li2 != -1)
                    drawHorizontalLine(li[li1], li[li2], y, buffer);
            }
        }
    }

}
